==> kom%2B%2B-0.5.tar.gz: kom++-0.5/libkom++/membership-cache.cc <==
// Cache of memberships.
//
// $Id: membership-cache.cc,v 1.6 1996/01/21 20:58:57 ceder Exp $
// Copyright (C) 1994  Per Cederqvist and Inge Wallin
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include <stdlib.h>
#include <assert.h>
#include "membership-cache.h"
#include "question.h"
#include "misc.h"
#include "error-codes.h"

static Membership bad_membership;


//FIXME: Extremely ad-hoc.  Is there a better way??
#define PNO_CNO_TO_INT(pno,cno) ((int)(pno)*10000+(int)(cno))

Membership_cache::Membership_cache(Connection *c)
{
    conn = c;
}

//Membership_cache::Membership_cache(const Membership_cache& c)
//{
//    *this = c;
//}

//Membership_cache &
//Membership_cache::operator=(const Membership_cache& c)
//{
//    cerr << "Membership_cache::operator=(const Membership_cache& c) unimp\n";
//    abort();
//}

Membership_cache::~Membership_cache()
{
}

Connection *
Membership_cache::connection() const
{
    return conn;
}

bool
Membership_cache::present(Pers_no pno, Conf_no cno) const
{
    return statusmap.present(PNO_CNO_TO_INT(pno, cno));
}

//void
//Membership_cache::clear()
//{
//    cerr << "Unimp\n";
//    abort();
//}


Status
Membership_cache::prefetch(Pers_no pno, Conf_no cno)
{
    int  index = PNO_CNO_TO_INT(pno, cno);

    if (index <= 0)
	return st_error;

    if (statusmap.present(index))
    {
	if (statusmap[index].valid())
	    return st_ok;
	else
	    return st_error;
    }
    else
    {
	if (!pendingmap.present(index))
	{
	    pendingmap[index] = new get_membership_question(conn, pno, cno);
	    return st_pending;
	}
	else
	{
	    switch (pendingmap[index]->status())
	    {
	    case st_pending:
		return st_pending;
	    case st_ok:
		statusmap[index] = pendingmap[index]->result();
		pendingmap.delete_entry(index);
		return st_ok;
	    case st_error:
 		statusmap[index] = Membership(pendingmap[index]->error());
		pendingmap.delete_entry(index);
		return st_error;
	    default:
		abort();
	    }
	}
    }
    abort();
}


Membership
Membership_cache::get(Pers_no pno, Conf_no cno)
{
    int  index = PNO_CNO_TO_INT(pno, cno);

    assert(index > 0);

    if (statusmap.present(index))
	return statusmap[index];
    else if (pendingmap.present(index))
    {
	get_membership_question *gtr = pendingmap[index];
	pendingmap.delete_entry(index);
	if (gtr->receive() == st_ok)
	    statusmap[index] = gtr->result();
	else
	    statusmap[index] = Membership(gtr->error());
	delete gtr;
	return statusmap[index];
    }
    else
    {
	// Not present in cache, and no outstanding question.
	get_membership_question gt(conn, pno, cno);
	if (gt.receive() == st_ok)
	    statusmap[index] = gt.result();
	else
	    statusmap[index] = Membership(gt.error());
	return statusmap[index];
    }
    abort();
}

//void
//Membership_cache::remove(Pers_no pno, Conf_no cno)
//{
//    abort();
//}

bool
Membership_cache::Immap::present(int index) const
{
    // FIXME:
    // contains() really should be a const member function. This is a 
    // workaround. The line below should be substituted.
    return (((Membership_cache::Immap*)this)->contains(index) ? true : false);
    // return contains(c) ? true : false;
}

Membership &
Membership_cache::Immap::operator[](int index)
{
    return intMembershipCHMap::operator[](index);
}


Membership_cache::Immap::Immap()
    : intMembershipCHMap(bad_membership, 50)
{
}

bool
Membership_cache::Imreqmap::present(int index)
{
    return contains(index) ? true : false;
}

void
Membership_cache::Imreqmap::delete_entry(int index)
{
    del(index);
}

get_membership_question *&
Membership_cache::Imreqmap::operator[](int index)
{
    return intget_membership_question_ptrCHMap::operator[](index);
}

static get_membership_question_ptr bad_ptr = NULL;

Membership_cache::Imreqmap::Imreqmap()
    : intget_membership_question_ptrCHMap(bad_ptr)
{
}

==> kom%2B%2B-0.6.tar.gz: kom++-0.6/libkom++/membership-cache.cc <==
// Cache of memberships.
//
// $Id: membership-cache.cc,v 1.9 1996/07/18 11:53:10 ceder Exp $
// Copyright (C) 1994  Per Cederqvist and Inge Wallin
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include <stdlib.h>
#include <assert.h>
#include "membership-cache.h"
#include "question.h"
#include "misc.h"
#include "error-codes.h"

static Membership bad_membership;


//FIXME: Extremely ad-hoc.  Is there a better way??
#define PNO_CNO_TO_INT(pno,cno) ((int)(pno)*10000+(int)(cno))

Membership_cache::Membership_cache(Connection *c)
{
    conn = c;
}

//Membership_cache::Membership_cache(const Membership_cache& c)
//{
//    *this = c;
//}

//Membership_cache &
//Membership_cache::operator=(const Membership_cache& c)
//{
//    cerr << "Membership_cache::operator=(const Membership_cache& c) unimp\n";
//    abort();
//}

Membership_cache::~Membership_cache()
{
}

Connection *
Membership_cache::connection() const
{
    return conn;
}

bool
Membership_cache::present(Pers_no pno, Conf_no cno) const
{
    return statusmap.present(PNO_CNO_TO_INT(pno, cno));
}

//void
//Membership_cache::clear()
//{
//    cerr << "Unimp\n";
//    abort();
//}


Status
Membership_cache::prefetch(Pers_no pno, Conf_no cno)
{
    int  index = PNO_CNO_TO_INT(pno, cno);

    if (index <= 0)
	return st_error;

    bool present = statusmap.present(index);
    
    if (present && statusmap[index].errno() != KOM_INVALIDATED)
    {
	if (statusmap[index].errno() == KOM_NO_ERROR)
	    return st_ok;
	else
	    return st_error;
    }
    else
    {
	if (!pendingmap.present(index))
	{
	    pendingmap[index] = new get_membership_question(conn, pno, cno);
	    return st_pending;
	}
	else
	{
	    switch (pendingmap[index]->status())
	    {
	    case st_pending:
		return st_pending;
	    case st_ok:
		if (present)
		    statusmap[index].set(pendingmap[index]->result());
		else
		    statusmap[index] = pendingmap[index]->result();
		pendingmap.delete_entry(index);
		return st_ok;
	    case st_error:
		if (present)
		    statusmap[index].set(
			Membership(pendingmap[index]->error()));
		else
		    statusmap[index] = Membership(pendingmap[index]->error());
		pendingmap.delete_entry(index);
		return st_error;
	    default:
		abort();
	    }
	}
    }
    abort();
}

bool
Membership_cache::prefetch_pending(Pers_no pno, Conf_no cno)
{
    int  index = PNO_CNO_TO_INT(pno, cno);

    if (index <= 0)
	return false;

    return (pendingmap.present(index)
	    && pendingmap[index]->status() == st_pending);
}


Membership
Membership_cache::get(Pers_no pno, Conf_no cno)
{
    int  index = PNO_CNO_TO_INT(pno, cno);
    bool present = false;

    assert(index > 0);

    if (statusmap.present(index))
    {
	Membership retval = statusmap[index];
	if (retval.errno() != KOM_INVALIDATED)
	    return retval;
	else
	    present = true;
    }

    if (pendingmap.present(index))
    {
	get_membership_question *gtr = pendingmap[index];
	pendingmap.delete_entry(index);
	Membership result;
	if (gtr->receive() == st_ok)
	    result = gtr->result();
	else
	    result = Membership(gtr->error());
	if (present) 
	{
	    statusmap[index].set(result);
	    result = statusmap[index];
	}
	else
	    statusmap[index] = result;
	delete gtr;
	return result;
    }
    else
    {
	// Not present in cache, and no outstanding question.
	get_membership_question gt(conn, pno, cno);
	Membership result;
	if (gt.receive() == st_ok)
	    result = gt.result();
	else
	    result = Membership(gt.error());
	if (present) 
	{
	    statusmap[index].set(result);
	    result = statusmap[index];
	}
	else
	    statusmap[index] = result;
	return result;
    }
    abort();
}

// FIXME: implementing this is tricky, since several places in kom++
// may hold Membership objects.  It should probably never be done,
// since they share data.  This (the cache) is the one place where a
// mapping from (pno,cno) to Membership is done -- it should always
// return the same Membership!  It is probably OK to remove the
// membership if the reference count is 1 (meaning that the cache
// holds the only copy).
//void
//Membership_cache::remove(Pers_no pno, Conf_no cno)
//{
//    abort();
//}

bool
Membership_cache::Immap::present(int index) const
{
    // FIXME:
    // contains() really should be a const member function. This is a 
    // workaround. The line below should be substituted.
    return (((Membership_cache::Immap*)this)->contains(index) ? true : false);
    // return contains(c) ? true : false;
}

Membership &
Membership_cache::Immap::operator[](int index)
{
    return intMembershipCHMap::operator[](index);
}


Membership_cache::Immap::Immap()
    : intMembershipCHMap(bad_membership, 50)
{
}

bool
Membership_cache::Imreqmap::present(int index)
{
    return contains(index) ? true : false;
}

void
Membership_cache::Imreqmap::delete_entry(int index)
{
    del(index);
}

get_membership_question *&
Membership_cache::Imreqmap::operator[](int index)
{
    return intget_membership_question_ptrCHMap::operator[](index);
}

static get_membership_question_ptr bad_ptr = NULL;

Membership_cache::Imreqmap::Imreqmap()
    : intget_membership_question_ptrCHMap(bad_ptr)
{
}
